from io import TextIOWrapper
from typing import Any, Callable, Optional, SupportsInt, Union, IO
from os import system as runsys, isatty
from time import time as epochTime, sleep
from inspect import getsourcelines

from . sets import CharSet, ColorSet, FormatSet, CharSetEntry, FormatSetEntry, ColorSetEntry
from . utils import *
from . cond import Cond


if not isatty(0):	# check if we are running in a tty
	raise RuntimeError("pbar requires a terminal emulator to work properly")
runsys("")		# We need to do this, otherwise Windows won't display special VT100 sequences


Position = tuple[Union[str, int], Union[str, int]]


def _genShape(position: tuple[int, int], size: tuple[int, int], charset: CharSet,
			  parsedColorset: dict, filled: Optional[str] = " ") -> str:
	"""Generates a basic rectangular shape that uses a charset and a parsed colorset"""
	width, height = size[0] + 2, size[1]

	charVert = (
		parsedColorset["vert"]["left"] + charset["vert"]["left"],
		parsedColorset["vert"]["right"] + charset["vert"]["right"]
	)
	charHoriz = (
		charset["horiz"]["top"],
		charset["horiz"]["bottom"],
	)
	charCorner = (
		parsedColorset["corner"]["tleft"] + charset["corner"]["tleft"],
		parsedColorset["corner"]["tright"] + charset["corner"]["tright"],
		parsedColorset["corner"]["bleft"] + charset["corner"]["bleft"],
		parsedColorset["corner"]["bright"] + charset["corner"]["bright"]
	)


	top: str = (
		Term.pos(position)
		+ charCorner[0]
		+ parsedColorset["horiz"]["top"] + charHoriz[0]*width
		+ charCorner[1]
	)

	mid: str = "".join((
		Term.pos(position, (0, row))
		+ charVert[0]
		+ (Term.moveHoriz(width) if filled is None else filled[0]*width)
		+ charVert[1]
	) for row in range(1, height))

	bottom: str = (
		Term.pos(position, (0, height))
		+ charCorner[2]
		+ parsedColorset["horiz"]["bottom"] + charHoriz[1]*width
		+ charCorner[3]
	)

	return top + mid + bottom


def _genBarContent(position: tuple[int, int], size: tuple[int, int], charset: CharSet,
				   parsedColorset: ColorSet, rangeValue: tuple[int, int]) -> str:
	"""Generates the progress shape with a parsed colorset and a charset specified"""
	width, height = size
	SEGMENTS_FULL = int((capValue(rangeValue[0], rangeValue[1], 0) / capValue(rangeValue[1], min=1))*width)	# Number of character for the full part of the bar
	SEGMENTS_EMPTY = width - SEGMENTS_FULL

	charFull = charset["full"]
	charEmpty = charset["empty"]

	return "".join((
			Term.pos((position), (0, row))
			+ parsedColorset["full"] + charFull*SEGMENTS_FULL
			+ parsedColorset["empty"] + charEmpty*SEGMENTS_EMPTY
		) for row in range(1, height))


def _genBarText(position: tuple[int, int], size: tuple[int, int],
				parsedColorset: dict[str, Union[dict, str]], formatset: FormatSet) -> str:
	"""Generates all text for the bar"""
	width, height = size

	def stripText(string: str, maxlen: int):
		"""Return a string stripped if the len of it is larger than the maxlen specified"""
		maxlen = capValue(maxlen, min=3)
		return string[:maxlen-3] + "..." if len(string) > maxlen else string

	txtMaxWidth = width + 2
	txtSubtitle = stripText(formatset["subtitle"], txtMaxWidth)
	txtInside = stripText(formatset["inside"], txtMaxWidth - 4)
	txtTitle = stripText(formatset["title"], txtMaxWidth)

	textTitle = (
		Term.pos(position, (-1, 0))
		+ parsedColorset["text"]["title"]
		+ txtTitle
	)

	textSubtitle = (
		Term.pos(position, (width - len(txtSubtitle) + 1, height))
		+ parsedColorset["text"]["subtitle"]
		+ txtSubtitle
	)

	textRight = (
		Term.pos(position, (width + 3, height/2))
		+ parsedColorset["text"]["right"]
		+ formatset["right"]
		+ Term.CLEAR_RIGHT
	) if formatset["right"] else ""

	textLeft = (
		Term.pos(position, (-len(formatset["left"]) - 3, height/2))
		+ Term.CLEAR_LEFT
		+ parsedColorset["text"]["left"]
		+ formatset["left"]
	) if formatset["left"] else ""

	txtInside = (
		Term.pos(position, (width/2 - len(txtInside)/2, height/2))
		+ parsedColorset["text"]["inside"]
		+ txtInside
	)

	return textTitle + textSubtitle + textRight + textLeft + txtInside




class PBar():
	"""Object for managing a progress bar."""
	def __init__(self,
			prange: tuple[int, int]=(0, 1), text: str=None, size: tuple[int, int]=(20, 1),
			position: tuple[Union[str, int], Union[str, int]]=("center", "center"),
			charset: Optional[CharSetEntry]=None, colorset: Optional[ColorSetEntry]=None,
			formatset: Optional[FormatSetEntry]=None, conditions: Optional[Union[list[Cond], Cond]]=None
		) -> None:
		"""
		### Detailed descriptions:
		@prange: This tuple will specify the range of two values to display in the progress bar.

		---

		@text: String to show in the progress bar.

		---

		@size: Tuple that specifies the width and height of the bar.

		---

		@position: Tuple containing the position (X and Y axles of the center) of the progress bar on the terminal.

		- If an axis value is `center`, the bar will be positioned at the center of the terminal on that axis.
		- Negative values will position the bar at the other side of the terminal.

		---

		@charset: Set of characters to use when drawing the progress bar.

		To use one of the included sets, use any of the constant values in `pbar.CharSet`. Keep in mind that some fonts might not have
		the characters used in some charsets.

		Since this value is just a dictionary, it is possible to use custom sets, which should specify the custom characters.

		---

		@colorset: Set of colors to use when drawing the progress bar.

		To use one of the included sets, use any of the constant values in `pbar.ColorSet`.

		Since this value is just a dictionary, it is possible to use custom sets, which should specify the custom colors.

		---

		@formatset: Formatting used when displaying the strings in different places around the bar.

		To use one of the included sets, use any of the constant values in `pbar.FormatSet`.

		Since this value is just a dictionary, it is possible to use custom sets, which should specify the custom formatting.

		---

		@conditions: One or more conditions to check before each time the bar draws.
		If one succeeds, the specified customization sets will be applied to the bar.

		To create a condition, use `pbar.Cond`.
		"""
		self._requiresClear = False		# This controls if the bar needs to clear its old position before drawing.
		self._enabled = True			# If disabled, the bar will never draw.
		self._time = epochTime()			# The elapsed time since the bar created.
		self._isOnScreen = False		# Is the bar on screen?

		self._range = PBar._getRange(prange)
		self._text = FormatSet._rmPoisonChars(text) if text is not None else ""
		self._formatset = FormatSet(formatset)
		self._size = PBar._getSize(size)
		self._charset = CharSet(charset)
		self._colorset = ColorSet(colorset)
		self._pos = self._getPos(position)
		self._conditions = PBar._getConds(conditions)

		self._oldValues = [self._pos, self._size]	# This values are used when clearing the old position of the bar (when self._requiresClear is True)


	# --------- Properties / Methods the user should use. ----------


	def draw(self):
		"""Print the progress bar on screen."""
		if self._requiresClear:
			# Clear the bar at the old position and length
			self._printStr(self._genClearedBar(self._oldValues))
			self._oldValues = [self._pos, self._size]	# Reset the old values

		# Draw the bar
		self._printStr(self._genBar())

		self._requiresClear = False


	def step(self, steps: int=1, text=None):
		"""
		Add `steps` to the first value in prange, then draw the bar.
		@steps: Value to add to the first value in prange.
		@text: Text to be displayed on the bar.
		"""
		self.prange = (self._range[0] + steps, self._range[1])
		if text is not None: self.text = text
		self.draw()


	def clear(self):
		"""Clear the progress bar."""
		bar = self._genClearedBar((self._pos, self._size))
		self._printStr(bar)


	def resetETime(self):
		"""Reset the elapsed time counter."""
		self._time = epochTime()	# Just set _time to the current time.


	@classmethod
	def fromConfig(cls, other: Union["PBar", dict]) -> "PBar":
		"""Return a PBar object with the new configuration given from another PBar or dict."""
		pb = cls()
		pb.config = other.config if isinstance(other, PBar) else other
		return pb


	def prangeFromFile(self, fp: IO[str]):
		"""Modify `prange` with the number of lines of a file."""
		chkInstOf(fp, TextIOWrapper, name="fp")
		self.prange = (0, len(fp.readlines()))
		fp.seek(0)


	@property
	def percentage(self):
		"""Percentage of the progress of the current `prange`."""
		return int((self._range[0]*100) / self._range[1])


	@property
	def text(self):
		"""Text to be displayed on the bar."""
		return self._text
	@text.setter
	def text(self, text: str):
		self._text = FormatSet._rmPoisonChars(text)


	@property
	def prange(self) -> tuple[int, int]:
		"""Range for the bar progress."""
		return (self._range[0], self._range[1])
	@prange.setter
	def prange(self, range: tuple[int, int]):
		self._range = PBar._getRange(range)


	@property
	def charset(self) -> CharSet:
		"""Set of characters for the bar."""
		return self._charset
	@charset.setter
	def charset(self, charset: CharSetEntry):
		self._charset = CharSet(charset)


	@property
	def colorset(self) -> ColorSet:
		"""Set of colors for the bar."""
		return self._colorset
	@colorset.setter
	def colorset(self, colorset: ColorSetEntry):
		self._colorset = ColorSet(colorset)


	@property
	def formatset(self) -> FormatSet:
		"""Formatting used for the bar."""
		return self._formatset
	@formatset.setter
	def formatset(self, formatset: FormatSetEntry):
		self._formatset = FormatSet(formatset)


	@property
	def size(self):
		"""Size of the progress bar."""
		return self._size
	@size.setter
	def size(self, size: tuple[int, int]):
		newsize = PBar._getSize(size)
		self._requiresClear = newsize != self._size
		self._oldValues[1] = self._size
		self._size = newsize


	@property
	def position(self):
		"""Position of the progress bar."""
		return self._pos
	@position.setter
	def position(self, position: Position):
		newpos = self._getPos(position)
		# we dont need to update the position and ask to redraw if the value supplied is the same
		if newpos != self._pos:
			self._oldValues[0] = self._pos
			self._requiresClear = True		# Position has been changed, we need to clear the bar at the old position
			self._pos = newpos


	@property
	def enabled(self):
		"""Will the bar draw on the next `draw` calls?"""
		return self._enabled
	@enabled.setter
	def enabled(self, state: bool):
		self._enabled = state


	@property
	def etime(self) -> float:
		"""Time elapsed since the bar created."""
		return round(epochTime() - self._time, 2)


	@property
	def conditions(self) -> tuple:
		"""Conditions for the bar."""
		return self._conditions
	@conditions.setter
	def conditions(self, conditions: Optional[Union[list[Cond], Cond]]):
		self._conditions = PBar._getConds(conditions)


	@property
	def config(self) -> dict:
		"""All the values of the progress bar stored in a dict."""
		return {
			"prange":		self._range,
			"text":			self._text,
			"size":			self._size,
			"position":		self._pos,
			"charset":		self._charset,
			"colorset":		convertClrs(self._colorset, "HEX"),
			"formatset":	self._formatset,
			"conditions":	self._conditions,
			"enabled":		self._enabled
		}
	@config.setter
	def config(self, config: dict[str, Any]):
		chkInstOf(config, dict, name="config")
		for key in {"prange", "text", "size", "position", "charset", "colorset", "formatset", "conditions", "enabled"}:
			# Iterate through every key in the dict and populate the config of the class with its values
			if key not in config:
				raise ValueError(f"config dict is missing the {key!r} key")
			setattr(self, key, config[key])


	# --------- ///////////////////////////////////////// ----------


	def _getPos(self, position: Position) -> tuple[int, int]:
		"""Get and process new position requested"""
		chkSeqOfLen(position, 2)

		TERM_SIZE: tuple[int, int] = Term.size()
		newpos = []

		for index, value in enumerate(position):
			if value == "center":
				value = int(TERM_SIZE[index]/2)+1
			elif value == "relative":
				value = Term.getPos()[index]
			chkInstOf(value, int, float, name="pos")

			if value < 0:
				value = TERM_SIZE[index] + value

			if index == 0:
				value = capValue(value, TERM_SIZE[0] - self._size[0]/2 - 1, self._size[0]/2 + 3)
			else:
				value = capValue(value, TERM_SIZE[1] - self._size[1]/2 - 1, self._size[1]/2 + 2)

			newpos.append(int(value))
		return (newpos[0], newpos[1])


	@staticmethod
	def _getSize(size: tuple[SupportsInt, SupportsInt]) -> tuple[int, int]:
		"""Get and process new length requested"""
		chkSeqOfLen(size, 2)
		width, height = map(int, size)
		return (capValue(int(width), min=5),
				capValue(int(height), min=1))


	@staticmethod
	def _getRange(range: tuple[SupportsInt, SupportsInt]) -> tuple[int, int]:
		"""Return a capped range"""
		chkSeqOfLen(range, 2)
		start, stop = map(int, range)
		return (capValue(start, stop, 0),
				capValue(stop, min=1))


	@staticmethod
	def _getConds(conditions: Union[list[Cond], Cond]) -> list[Cond]:
		if isinstance(conditions, (tuple, list)):
			for item in conditions:	chkInstOf(item, Cond)
			return conditions
		elif isinstance(conditions, Cond):
			return (conditions, )
		else:
			return ()


	def _chkConds(self) -> None:
		for cond in self._conditions:
			if not cond.test(self):
				continue
			if cond.newSets[0]:	self.charset = cond.newSets[0]
			if cond.newSets[1]:	self.colorset = cond.newSets[1]
			if cond.newSets[2]:	self.formatset = cond.newSets[2]




	def _genClearedBar(self, values: tuple[tuple[int, int], tuple[int, int]]) -> str:
		"""Generate a cleared progress bar. `values[0]` is the position, and `values[1]` is the size"""
		if not self._isOnScreen:	return ""
		pos, size = values
		parsedColorSet = ColorSet.parsedValues(ColorSet.EMPTY)

		size = size[0], size[1] + 1
		POSITION = (pos[0] + int(size[0]/-2),
					pos[1] + int(size[1]/-2))

		barShape = _genShape(
			(POSITION[0] - 2, POSITION[1]),
			size,
			CharSet.EMPTY,
			parsedColorSet
		)

		barText = _genBarText(
			POSITION,
			size,
			parsedColorSet,
			FormatSet.cleanedValues(self._formatset.parsedValues(self))
		)

		self._isOnScreen = False
		return barText + barShape


	def _genBar(self) -> str:
		"""Generate the progress bar"""
		if self._conditions:	self._chkConds()
		size = self._size[0], self._size[1] + 1
		POSITION = (self._pos[0] + int(size[0]/-2),
					self._pos[1] + int(size[1]/-2))

		parsedColorSet = self._colorset.parsedValues()

		# Build all the parts of the progress bar
		barShape = _genShape(
			(POSITION[0] - 2, POSITION[1]),
			size, self._charset, parsedColorSet
		)

		barContent = _genBarContent(
			POSITION,
			size, self._charset, parsedColorSet, self._range
		)

		barText = _genBarText(
			POSITION,
			size, parsedColorSet, self._formatset.parsedValues(self)
		)

		self._isOnScreen = True
		return barShape + barContent + barText


	def _printStr(self, barString: str):
		"""Prints string to stream"""
		if not self._enabled: return
		print(
			Term.CURSOR_SAVE + Term.CURSOR_HIDE
			+ barString
			+ Term.CURSOR_LOAD + Term.CURSOR_SHOW,
			flush=True,
			end=""
		)




def taskWrapper(barObj: PBar, scope: dict, titleComments=False, overwriteRange=True) -> Callable:
	"""
	Use as a decorator. Takes a PBar object, sets its prange depending on the quantity of
	statements inside the decorated function, and `steps` the bar over after every function statement.
	Note: Multi-line expressions are not supported.

	@barObj: PBar object to use.
	@scope: Dictionary containing the scope local variables.
	@titleComments: If True, comments on a statement starting with "#bTitle:" will be treated as titles for the progress bar.
	@overwriteRange: If True, overwrites the prange of the bar.
	"""
	chkInstOf(barObj, PBar, name="pbarObj")

	def getTitleComment(string: str) -> Optional[str]:
		"""Returns the text after "#bTitle:" from the string supplied. Returns None if there is no comment."""
		try:
			index = string.rindex("#bTitle:") + 8	# lol
		except ValueError:
			return ""
		return string[index:].strip()

	def wrapper(func):
		lines = getsourcelines(func)[0][2:]	# Get the source lines of code

		def wrapper2():
			barObj.prange = (0, len(lines)) if overwriteRange else barObj.prange

			for inst in lines:	# Iterate through every statement
				instComment = getTitleComment(inst)
				if titleComments and instComment:	barObj.text = instComment
				barObj.draw()
				try:
					eval(inst, scope)	# yep, this uses evil()
				except SyntaxError:
					raise RuntimeError("Multi-line expressions are not supported inside functions decorated with taskWrapper")
				barObj.step()
		return wrapper2
	return wrapper


def animate(barObj: PBar, rng: range, delay: float=0.1) -> None:
	"""
	Animates the given PBar object by filling it by the range given, with a delay.
	@barObj: PBar object to use.
	@rng: range object to set for the bar.
	@delay: Time in seconds between each time the bar draws.
	"""
	chkInstOf(rng, range, name="rng")
	steps = rng.step
	barObj.prange = rng.start, rng.stop
	for _ in rng:
		barObj.step(steps)
		sleep(delay)


def barHelper(position: Position=("center", "center"),
			  size: tuple[int, int]=(20, 1)) -> Position:
	"""
	Draw a bar helper on screen indefinitely until the user exits.
	Returns the position of the bar helper.
	@position: Position of the helper on the terminal.
	@size: Size of the helper.
	"""
	b = PBar(	# we create a bar just to make stuff easier
		size=size,
		formatset=FormatSet.EMPTY,
		colorset={
			"vert": "#090",
			"horiz": "#090",
			"corner": "#090",
			"text":	(255, 100, 0),
			"empty": "#090"
		}
	)

	print(Term.BUFFER_NEW + Term.CURSOR_HIDE)

	try:
		while True:
			b.position = position
			rPos = b.position
			b.formatset = {"subtitle": f"X:{rPos[0]} Y:{rPos[1]}"}

			xLine = Term.pos((0, rPos[1])) + "═"*rPos[0]
			yLine = "".join(Term.pos((rPos[0], x)) + "║" for x in range(rPos[1]))
			center = Term.pos(rPos) + "╝"

			print(
				Term.CLEAR_ALL
				+ b._genBar()
				+ Term.color((255, 100, 0)) + xLine + yLine + center
				+ Term.CURSOR_HOME + Term.INVERT + "Press Ctrl-C to exit." + Term.RESET,
				end=""
			)
			sleep(0.01)
	except KeyboardInterrupt:
		pass

	print(Term.BUFFER_OLD + Term.CURSOR_SHOW, end="")
	del b	# delete the bar we created
	return rPos


from typing import SupportsFloat, TypeVar, Optional, Union, Any, SupportsInt
from os import get_terminal_size

__all__ = (
	"capValue", "convertClrs", "chkInstOf",
	"chkSeqOfLen", "isNum", "Term"
)



_HTML_COLOR_NAMES: dict = {		# thanks to https://stackoverflow.com/a/1573141/14546524
	"aliceblue":"#f0f8ff", "antiquewhite":"#faebd7", "aqua":"#00ffff", "aquamarine":"#7fffd4", "azure":"#f0ffff",
	"beige":"#f5f5dc", "bisque":"#ffe4c4", "black":"#000000", "blanchedalmond":"#ffebcd", "blue":"#0000ff", "blueviolet":"#8a2be2", "brown":"#a52a2a", "burlywood":"#deb887",
	"cadetblue":"#5f9ea0", "chartreuse":"#7fff00", "chocolate":"#d2691e", "coral":"#ff7f50", "cornflowerblue":"#6495ed", "cornsilk":"#fff8dc", "crimson":"#dc143c", "cyan":"#00ffff",
	"darkblue":"#00008b", "darkcyan":"#008b8b", "darkgoldenrod":"#b8860b", "darkgray":"#a9a9a9", "darkgreen":"#006400", "darkkhaki":"#bdb76b", "darkmagenta":"#8b008b", "darkolivegreen":"#556b2f",
	"darkorange":"#ff8c00", "darkorchid":"#9932cc", "darkred":"#8b0000", "darksalmon":"#e9967a", "darkseagreen":"#8fbc8f", "darkslateblue":"#483d8b", "darkslategray":"#2f4f4f", "darkturquoise":"#00ced1",
	"darkviolet":"#9400d3", "deeppink":"#ff1493", "deepskyblue":"#00bfff", "dimgray":"#696969", "dodgerblue":"#1e90ff",
	"firebrick":"#b22222", "floralwhite":"#fffaf0", "forestgreen":"#228b22", "fuchsia":"#ff00ff",
	"gainsboro":"#dcdcdc", "ghostwhite":"#f8f8ff", "gold":"#ffd700", "goldenrod":"#daa520", "gray":"#808080", "green":"#008000", "greenyellow":"#adff2f",
	"honeydew":"#f0fff0", "hotpink":"#ff69b4",
	"indianred":"#cd5c5c", "indigo":"#4b0082", "ivory":"#fffff0", "khaki":"#f0e68c",
	"lavender":"#e6e6fa", "lavenderblush":"#fff0f5", "lawngreen":"#7cfc00", "lemonchiffon":"#fffacd", "lightblue":"#add8e6", "lightcoral":"#f08080", "lightcyan":"#e0ffff", "lightgoldenrodyellow":"#fafad2",
	"lightgrey":"#d3d3d3", "lightgreen":"#90ee90", "lightpink":"#ffb6c1", "lightsalmon":"#ffa07a", "lightseagreen":"#20b2aa", "lightskyblue":"#87cefa", "lightslategray":"#778899", "lightsteelblue":"#b0c4de",
	"lightyellow":"#ffffe0", "lime":"#00ff00", "limegreen":"#32cd32", "linen":"#faf0e6",
	"magenta":"#ff00ff", "maroon":"#800000", "mediumaquamarine":"#66cdaa", "mediumblue":"#0000cd", "mediumorchid":"#ba55d3", "mediumpurple":"#9370d8", "mediumseagreen":"#3cb371", "mediumslateblue":"#7b68ee",
	"mediumspringgreen":"#00fa9a", "mediumturquoise":"#48d1cc", "mediumvioletred":"#c71585", "midnightblue":"#191970", "mintcream":"#f5fffa", "mistyrose":"#ffe4e1", "moccasin":"#ffe4b5",
	"navajowhite":"#ffdead", "navy":"#000080",
	"oldlace":"#fdf5e6", "olive":"#808000", "olivedrab":"#6b8e23", "orange":"#ffa500", "orangered":"#ff4500", "orchid":"#da70d6",
	"palegoldenrod":"#eee8aa", "palegreen":"#98fb98", "paleturquoise":"#afeeee", "palevioletred":"#d87093", "papayawhip":"#ffefd5", "peachpuff":"#ffdab9", "peru":"#cd853f", "pink":"#ffc0cb", "plum":"#dda0dd", "powderblue":"#b0e0e6", "purple":"#800080",
	"rebeccapurple":"#663399", "red":"#ff0000", "rosybrown":"#bc8f8f", "royalblue":"#4169e1",
	"saddlebrown":"#8b4513", "salmon":"#fa8072", "sandybrown":"#f4a460", "seagreen":"#2e8b57", "seashell":"#fff5ee", "sienna":"#a0522d", "silver":"#c0c0c0", "skyblue":"#87ceeb", "slateblue":"#6a5acd", "slategray":"#708090", "snow":"#fffafa", "springgreen":"#00ff7f", "steelblue":"#4682b4",
	"tan":"#d2b48c", "teal":"#008080", "thistle":"#d8bfd8", "tomato":"#ff6347", "turquoise":"#40e0d0",
	"violet":"#ee82ee",
	"wheat":"#f5deb3", "white":"#ffffff", "whitesmoke":"#f5f5f5",
	"yellow":"#ffff00", "yellowgreen":"#9acd32"
}


Num = TypeVar("Num", int, float)

def capValue(value: Num, max: Optional[Num]=None, min: Optional[Num]=None) -> Num:
	"""Clamp a value to a minimum and/or maximum value."""
	if max is not None and value > max:
		return max
	elif min is not None and value < min:
		return min
	else:
		return value


def convertClrs(clr: Union[dict[Any, Union[str, tuple]], str, tuple], type: str) -> Union[str, tuple, dict, None]:
	"""
	Convert color values to HEX and vice-versa
	@clr:	Color value to convert.
	@type:	Type of conversion to do ('RGB' or 'HEX')
	"""
	if isinstance(clr, dict):
		return {key: convertClrs(value, type) for key, value in clr.items()}
	elif isinstance(clr, str):
		color = clr.lower()
		if color in _HTML_COLOR_NAMES:	color = _HTML_COLOR_NAMES[color]
	else:
		color = clr

	if type == "RGB":
		if not isinstance(color, str) or not color.startswith("#"):
			return color

		clrs = color.lstrip("#")

		if len(clrs) == 3:
			clrs = "".join(c*2 for c in clrs)

		try:
			return tuple(int(clrs[i:i+2], 16) for i in (0, 2, 4))
		except ValueError:
			return color
	elif type == "HEX":
		if not isinstance(color, (tuple, list)) or len(color) != 3: return color

		capped = tuple(capValue(value, 255, 0) for value in color)
		return f"#{capped[0]:02x}{capped[1]:02x}{capped[2]:02x}"
	else:
		return ""


def chkSeqOfLen(obj: Any, length: int, name: str=None) -> bool:
	"""
	Check if an object is a Sequence and has the length specified.
	If fails, raises exception (`Sequence obj | name must have len items`).
	"""
	chkInstOf(obj, tuple, list)
	if len(obj) != length:
		raise ValueError(
			(name or f"Sequence {Term.color((255, 150, 0))}{obj!r}{Term.RESET}")
			+ " must have " + Term.color((0, 255, 0)) + str(length) + Term.RESET + " items, "
			+ "not " + Term.color((255, 0, 0)) + str(len(obj))
		)
	return True


def chkInstOf(obj: Any, *typ: Any, name: str=None) -> bool:
	"""
	Check if an object is an instance of any of the other objects specified.
	If fails, raises exception (`Value | name must be *typ, not obj`).
	"""
	if not isinstance(obj, typ):
		raise TypeError(
			(name or f"Value {Term.color((255, 150, 0))}{obj!r}{Term.RESET}")
			+ " must be " + ' or '.join(Term.color((0, 255, 0)) + x.__name__ + Term.RESET for x in typ)
			+ ", not " + Term.color((255, 0, 0)) + obj.__class__.__name__ + Term.RESET
		)
	return True


def isNum(obj: SupportsFloat) -> bool:
	"""Return True if `obj` can be casted to float."""
	try:
		float(obj)
	except ValueError:
		return False
	return True




class Term:
	"""Class for using terminal sequences a bit easier"""
	@staticmethod
	def size() -> tuple[int, int]:
		"""Get size of the terminal. Columns and rows."""
		return tuple(get_terminal_size())

	@staticmethod
	def pos(pos: tuple[SupportsInt, SupportsInt],
			offset: tuple[SupportsInt, SupportsInt]=(0, 0)) -> str:
		"""
		Position of the cursor on the terminal.
		@pos: Tuple containing the X and Y position values.
		@offset: Offset applied to `pos`. (Can be negative)
		"""
		chkSeqOfLen(pos, 2)

		position = (int(pos[0]) + int(offset[0]),
					int(pos[1]) + int(offset[1]))

		return f"\x1b[{position[1]};{position[0]}f"


	@staticmethod
	def color(rgb: Optional[tuple[int, int, int]], bg: bool=False) -> str:
		"""
		Color of the cursor.
		@rgb:	Tuple with three values from 0 to 255. (RED GREEN BLUE)
		@bg:	This color will be displayed on the background
		"""
		if not isinstance(rgb, (tuple, list)):
			return Term.RESET
		elif len(rgb) != 3:
			raise ValueError("Sequence must have 3 items")

		crgb = [capValue(value, 255, 0) for value in rgb]

		type = 48 if bg else 38
		return f"\x1b[{type};2;{crgb[0]};{crgb[1]};{crgb[2]}m"


	@staticmethod
	def moveHoriz(dist: SupportsInt) -> str:
		"""Move the cursor horizontally `dist` characters (supports negative numbers)."""
		dist = int(dist)
		return f"\x1b[{abs(dist)}{'D' if dist < 0 else 'C'}"


	@staticmethod
	def moveVert(dist: SupportsInt, cr: bool=False) -> str:
		"""
		Move the cursor vertically `dist` lines (supports negative numbers).
		@cr: Do a carriage return too.
		"""
		dist = int(dist)
		return (
			f"\x1b[{abs(dist)}{'A' if dist < 0 else 'B'}"
			if not cr else
			f"\x1b[{abs(dist)}{'F' if dist < 0 else 'E'}"
		)


	@staticmethod
	def clear() -> str:
		"""
		Shortcut for `Term.CLEAR_ALL`, `Term.CLEAR_SCROLL` and `Term.CURSOR_HOME`.
		(Functionally the same as the `clear` command.)
		"""
		return Term.CURSOR_HOME + Term.CLEAR_ALL + Term.CLEAR_SCROLL


	@staticmethod
	def margin(top: SupportsInt=None, bottom: SupportsInt=None) -> str:
		"""Sets the top and bottom margins of the terminal. Use `None` for default margins."""
		return (
			Term.CURSOR_SAVE
			+ f"\x1b[{(int(top) + 1) if top else ''};{(Term.size()[1] - int(bottom)) if bottom else ''}r"
			+ Term.CURSOR_LOAD
		)


	@staticmethod
	def scroll(dist: SupportsInt) -> str:
		"""Scrolls the terminal buffer `dist` lines (supports negative numbers)."""
		dist = int(dist)
		return f"\x1b[{abs(dist)}{'T' if dist < 0 else 'S'}"


	@staticmethod
	def fill(char: str) -> str:
		"""Fill the terminal screen with the character specified."""
		ts = Term.size()
		return (
			Term.CURSOR_HOME
			+ "".join(Term.pos((0, row)) + char[0]*ts[0] for row in range(ts[1] + 1))
		)


	@staticmethod
	def formatString(string: str, reset: bool=True) -> str:  # sourcery no-metrics
		"""
		Add format to the string supplied by wrapping text with special characters:

		- `*`: Bold.
		- `_`: Italic.
		- `~`: Strikethrough.
		- `-`: Underline.
		- `´`: Blink.
		- `|`: Invert.
		- `#`: Dim.
		- `@`: Invisible.

		Note: Some of this sequences might not work properly on some terminal emulators.

		Note: When disabling `Dim`, bold will also be disabled.

		@reset: Will formatting be resetted at the end?
		"""
		invert = underline = dim = sthrough = invisible = bold = italic = blink = False
		ignoreChar = False
		endStr = ""
		for char in string:
			if char == "\\":
				ignoreChar = True
				continue
			if ignoreChar:
				endStr += char
				ignoreChar = False
				continue

			if char == "*":	# bold
				char = Term.NO_BOLD if bold else Term.BOLD
				bold = not bold
			elif char == "_":	# italic
				char = Term.NO_ITALIC if italic else Term.ITALIC
				italic = not italic
			elif char == "~":	# strikethrough
				char = Term.NO_STHROUGH if sthrough else Term.STHROUGH
				sthrough = not sthrough
			elif char == "-":	# underline
				char = Term.NO_UNDERLINE if underline else Term.UNDERLINE
				underline = not underline
			elif char == "´":	# blink
				char = Term.NO_BLINK if blink else Term.BLINK
				blink = not blink
			elif char == "|":	# invert
				char = Term.NO_INVERT if invert else Term.INVERT
				invert = not invert
			elif char == "#":	# dim
				char = Term.NO_DIM if dim else Term.DIM
				dim = not dim
			elif char == "@":	# invisible
				char = Term.NO_INVISIBLE if invisible else Term.INVISIBLE
				invisible = not invisible
			endStr += char
		return endStr + (Term.RESET if reset else "")


	# simple sequences that dont require parsing
	# text formatting
	INVERT: str =		"\x1b[7m"
	NO_INVERT: str =	"\x1b[27m"
	UNDERLINE: str =	"\x1b[4m"
	NO_UNDERLINE: str =	"\x1b[24m"
	DIM: str =			"\x1b[2m"
	NO_DIM: str =		"\x1b[22m"
	STHROUGH: str = 	"\x1b[9m"
	NO_STHROUGH: str = 	"\x1b[29m"
	INVISIBLE: str =	"\x1b[8m"
	NO_INVISIBLE: str =	"\x1b[28m"
	BOLD: str =			"\x1b[1m"
	NO_BOLD: str =		"\x1b[22m"
	ITALIC: str =		"\x1b[3m"
	NO_ITALIC: str =	"\x1b[23m"
	BLINK: str =		"\x1b[5m"
	NO_BLINK: str =		"\x1b[25m"
	RESET: str =		"\x1b[0m"

	# special
	CLEAR_LINE: str =	"\x1b[2K"
	CLEAR_RIGHT: str =	"\x1b[0K"
	CLEAR_LEFT: str =	"\x1b[1K"
	CLEAR_DOWN: str =	"\x1b[0J"
	CLEAR_ALL: str =	"\x1b[2J"
	CLEAR_SCROLL: str =	"\x1b[3J"
	CURSOR_SHOW: str =	"\x1b[?25h"
	CURSOR_HIDE: str =	"\x1b[?25l"
	CURSOR_SAVE: str =	"\x1b7"
	CURSOR_LOAD: str =	"\x1b8"
	BUFFER_NEW: str =	"\x1b[?1049h"
	BUFFER_OLD: str =	"\x1b[?1049l"
	CURSOR_HOME: str =	"\x1b[H"


from shlex import split as strSplit

from . utils import chkInstOf, chkSeqOfLen, isNum
from . import bar
from . sets import ColorSetEntry, FormatSetEntry, CharSetEntry, FormatSet

_OP_EQU = "=="
_OP_NEQ = "!="
_OP_GTR = ">"
_OP_LSS = "<"
_OP_GEQ = ">="
_OP_LEQ = "<="
_OP_IN = "<-"


class Cond:
	"""Condition manager used by a PBar object."""
	def __init__(self, condition: str, charset: CharSetEntry = None,
				 colorset: ColorSetEntry = None, formatset: FormatSetEntry = None) -> None:
		"""
		Apply different customization sets to a bar if the condition supplied succeeds.
		Text comparisons are case insensitive.

		The condition string must be composed of three values separated by spaces:

		1. Attribute key (Formatting keys for `pbar.FormatSet`)
		2. Comparison operator (`==`, `!=`, `>`, `<`, `>=`, `<=`, `<-`)
		3. Value

		- Note: The "custom" operator `<-` stands for the attribute key containing the value.

		---

		### Examples:

		>>> Cond("percentage >= 50", colorset=ColorSet.DARVIL)

		>>> Cond("text <- 'error'", colorset=ColorSet.ERROR, formatset=FormatSet.TITLE_SUBTITLE)
		"""
		vs = self._chkCond(condition)
		self._attribute, self.operator = vs[0:2]
		self._value = float(vs[2]) if isNum(vs[2]) else vs[2].lower()
		self.newSets = (charset, colorset, formatset)


	@staticmethod
	def _chkCond(cond: str):
		chkInstOf(cond, str, name="condition")
		splitted = strSplit(cond)
		chkSeqOfLen(splitted, 3, "condition")

		if splitted[1] not in {_OP_EQU, _OP_NEQ, _OP_GTR, _OP_LSS,
							   _OP_GEQ, _OP_LEQ, _OP_IN}:
			raise RuntimeError(f"Invalid operator {splitted[1]!r}")

		return splitted


	def __repr__(self) -> str:
		"""Returns `Cond('attrib operator value', *newSets)`"""
		return (f"{self.__class__.__name__}('{self._attribute} {self.operator} {self._value}', {self.newSets})")


	def test(self, cls: "bar.PBar") -> bool:
		"""Check if the condition succeededs with the values of the PBar object"""
		op = self.operator
		val = FormatSet._getBarAttrs(cls, self._attribute)
		val = val.lower() if isinstance(val, str) else val

		if op == _OP_EQU:
			return val == self._value
		elif op == _OP_NEQ:
			return val != self._value
		elif op == _OP_GTR:
			return val > self._value
		elif op == _OP_LSS:
			return val < self._value
		elif op == _OP_GEQ:
			return val >= self._value
		elif op == _OP_LEQ:
			return val <= self._value
		elif op == _OP_IN:
			return self._value in val
		else:
			return False